from apscheduler.schedulers.background import BackgroundScheduler  # type: ignore


def shutdown_scheduler(scheduler: BackgroundScheduler) -> None:
    if len(scheduler.get_jobs()) > 0:
        scheduler.remove_all_jobs()
    if scheduler.running:
        scheduler.shutdown()



import time
from sir_utilities.thread_helper import find_thread_by_name


def keep_main_alive_for_thread(thread_name: str) -> None:
    if thread := find_thread_by_name(f"{thread_name}"):
        while thread.is_alive():
            time.sleep(5)



from datetime import datetime
from enum import Enum, auto
import os
import logging
from sir_utilities.find_root import find_root


def create_log_files_dir(project_root_path: str = find_root()) -> str:
    log_files_dir = os.path.join(project_root_path, "log_files")

    # Check if the directory exists - if not create
    if not os.path.exists(log_files_dir):
        os.makedirs(log_files_dir)

    return log_files_dir


class LogMode(Enum):
    DEBUG = auto()
    INFO = auto()
    DEBUG_PRINT_ONLY = auto()
    INFO_PRINT_ONLY = auto()


def setup_logging(mode: LogMode, log_file_dir: str = create_log_files_dir()) -> None:
    default_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    today = datetime.now().strftime("%Y-%m-%d")

    if mode == LogMode.DEBUG:
        log_path = os.path.join(log_file_dir, f"debug_log_{today}.log")
        logging.basicConfig(
            force=True,
            level=logging.DEBUG,
            handlers=[logging.FileHandler(log_path), logging.StreamHandler()],
            format=default_format,
        )
    elif mode == LogMode.INFO:
        log_path = os.path.join(log_file_dir, f"live_log_{today}.log")
        logging.basicConfig(
            force=True,
            level=logging.INFO,
            handlers=[logging.FileHandler(log_path), logging.StreamHandler()],
            format=default_format,
        )
    elif mode == LogMode.DEBUG_PRINT_ONLY:
        logging.basicConfig(
            force=True,
            level=logging.DEBUG,
            handlers=[logging.StreamHandler()],
            format=default_format,
        )
    elif mode == LogMode.INFO_PRINT_ONLY:
        logging.basicConfig(
            force=True,
            level=logging.INFO,
            handlers=[logging.StreamHandler()],
            format=default_format,
        )


def module_logger(module_name: str) -> logging.Logger:
    logger = logging.getLogger(module_name)
    logger.propagate = True
    return logger



import os


def find_root(marker_file: str = "pyproject.toml") -> str:
    # Traverse project structure upwards until pyproject.toml is found
    temp_dir = start_dir = os.path.dirname(os.path.abspath(__file__))
    while temp_dir != "/":
        if _contains_marker_file(temp_dir, marker_file):
            return temp_dir
        temp_dir = os.path.dirname(temp_dir)
    raise FileNotFoundError(
        f"Project root not found (start: '{start_dir}', marker: '{marker_file}')."
    )


def _contains_marker_file(directory_path: str, marker_file: str) -> bool:
    pyproject_path = os.path.join(directory_path, marker_file)
    return os.path.exists(pyproject_path)



from datetime import datetime, timezone
from dateutil import parser
from enum import Enum


class DateFormat(Enum):
    YMDHMSF_ISO = "%Y-%m-%dT%H:%M:%S.%fZ"
    YMDHMSF_ISO_OFFSET = "%Y-%m-%d %H:%M:%S%z"
    YMDHM = "%Y-%m-%d %H:%M"


def now_utc_iso() -> datetime:
    return datetime.now(timezone.utc)


def string_to_datetime(
    date_str: str,
) -> datetime:
    dt = parser.isoparse(date_str)
    return dt.astimezone(tz=timezone.utc)


def datetime_to_string(
    date: datetime,
    format: str = DateFormat.YMDHMSF_ISO_OFFSET.value,
) -> str:
    return date.strftime(format)



from abc import ABC
from datetime import datetime, timedelta
from logging import Logger
import threading
import time

import psutil

from sir_utilities.date_time import now_utc_iso
from sir_utilities.logger import module_logger


class SystemMetricsLogger(ABC):
    def log_cpu_usage(self, logger: Logger, cpu_usage: str) -> None:
        msg = f"CPU usage: {cpu_usage}%."
        logger.info(msg)

    def log_memory_usage(
        self,
        logger: Logger,
        memory_total: float,
        memory_used: float,
        used_pct: float,
    ) -> None:
        mem_used_mb = f"{memory_used / (1024 ** 2):.2f}MB"
        mem_total_mb = f"{memory_total / (1024 ** 2):.2f}MB"
        percentage = f"{used_pct:.2f}%"
        usage_str = f"{mem_used_mb} / {mem_total_mb} ({percentage})"
        msg = f"RAM usage: {usage_str}."
        logger.info(msg)

    def log_swap_usage(
        self,
        logger: Logger,
        swap_total: float,
        swap_used: float,
        used_pct: float,
    ) -> None:
        swap_used_mb = f"{swap_used / (1024 ** 2):.2f}MB"
        swap_total_mb = f"{swap_total / (1024 ** 2):.2f}MB"
        percentage = f"{used_pct:.2f}%"
        usage_str = f"{swap_used_mb} / {swap_total_mb} ({percentage})"
        msg = f"Swap usage: {usage_str}."
        logger.info(msg)


class SystemMetrics(SystemMetricsLogger):
    def __init__(self) -> None:
        self._name = type(self).__name__
        self._logger = module_logger(type(self).__name__)
        self._running = False
        self._last_metrics: datetime | None = None

    def start(self, interval_seconds: int = 10) -> None:
        self._running = True
        t = threading.Thread(
            name=self._name,
            target=self.monitor_system_metrics,
            args=[interval_seconds],
        )
        t.start()

    def stop(self) -> None:
        self._running = False

    def monitor_system_metrics(self, interval_seconds: int):
        while self._running:
            if self._metrics_are_due(interval_seconds):
                self._cpu_usage()
                self._memory_usage()
                self._swap_usage()
                self._last_metrics = now_utc_iso()
            time.sleep(min(15, interval_seconds))

    def _metrics_are_due(self, interval_seconds: int) -> bool:
        if self._last_metrics is None:
            return True

        return now_utc_iso() - self._last_metrics > timedelta(seconds=interval_seconds)

    def _cpu_usage(self) -> None:
        cpu_usage_pct = str(psutil.cpu_percent())
        self.log_cpu_usage(self._logger, cpu_usage_pct)

    def _memory_usage(self) -> None:
        # Memory usage
        memory = psutil.virtual_memory()

        # Metrics
        memory_total = memory.total
        memory_used = memory.used
        used_pct = (memory_used / memory_total) * 100

        self.log_memory_usage(self._logger, memory_total, memory_used, used_pct)

    def _swap_usage(self) -> None:
        # Memory usage
        swap_memory = psutil.swap_memory()

        # Metrics
        swap_total = swap_memory.total
        swap_used = swap_memory.used
        used_pct = (swap_used / swap_total) * 100 if swap_total > 0 else float(0)

        self.log_swap_usage(self._logger, swap_total, swap_used, used_pct)



from datetime import datetime, timedelta
import threading
import time
from typing import Any, Callable, Dict


class Cli:
    def __init__(self, stop_sequence: str = "stop") -> None:
        self._stop_sequence = stop_sequence
        self._running = False

    def running(self) -> bool:
        return self._running

    def start_cli(
        self,
        duration_seconds: int,
        callbacks: Dict[Callable[..., None], Any | None],
    ) -> None:
        self._running = True
        self._callbacks = callbacks
        t_count_down = threading.Thread(
            target=self._count_down,
            args=[duration_seconds],
            name="cli_count_down",
        )
        t_cli = threading.Thread(target=self.listen_for_input, name="cli_input")
        t_count_down.start()
        t_cli.start()

    def stop_cli(self) -> None:
        if self._running:
            self._running = False
            for callback, args in self._callbacks.items():
                if args:
                    callback(args)
                else:
                    callback()

    def _count_down(self, duration: int | None) -> None:
        if duration:
            start_time = datetime.now()
            stop_time = start_time + timedelta(seconds=duration)

            while self._running and not datetime.now() >= stop_time:
                time.sleep(5)

            elapsed = datetime.now() - start_time
            elapsed = timedelta(days=elapsed.days, seconds=elapsed.seconds)

            print(f"Bot ran for {elapsed} hours.")
            self.stop_cli()

    def listen_for_input(self) -> None:
        while self._running:
            user_input = input(f"\nEnter '{self._stop_sequence}' to stop the bot: ")
            if user_input.lower() == f"{self._stop_sequence}":
                break
        self.stop_cli()



import threading


def find_thread_by_name(thread_name: str) -> threading.Thread | None:
    return next(
        (thread for thread in threading.enumerate() if thread.name == thread_name),
        None,
    )



from dotenv import load_dotenv
import os


def get_env_variable(variable: str, group: str | None = None) -> str:
    load_dotenv()
    if group:
        grouped_variable = f"{variable}_{group}"
        if result := os.getenv(grouped_variable):
            return result
        else:
            raise NameError(f"Variable {grouped_variable} not found.")

    elif result := os.getenv(variable):
        return result
    else:
        raise NameError(f"Variable {variable} not found.")



from typing import List, Union


def convert_bytes(size_bytes: Union[int, float]) -> str:
    units: List[str] = ["B", "KiB", "MiB", "GiB", "TiB"]
    unit_index: int = 0

    # Iterate through units until the size is small enough
    while size_bytes >= 1024 and unit_index < len(units) - 1:
        size_bytes /= 1024
        unit_index += 1

    # Format size with unit
    return f"{size_bytes:.2f} {units[unit_index]}"



from abc import ABC
from datetime import datetime, timedelta
from logging import Logger
import threading
import time
import tracemalloc
from typing import Any, List

# import gc
# from pympler.asizeof import asizeof  # type: ignore

from sir_utilities.date_time import now_utc_iso
from sir_utilities.logger import module_logger
from sir_utilities.unit_conversion import convert_bytes


class MemoryTracerLogger(ABC):
    def log_memory_allocations(
        self,
        logger: Logger,
        top_n: int,
        top_statistics: List[tracemalloc.StatisticDiff] | list[tracemalloc.Statistic],
    ) -> None:
        sorted_stats = sorted(
            top_statistics[:top_n],
            key=lambda stat: stat.size,
            reverse=True,
        )
        message = [f"Top {top_n} memory allocations by size:"]
        message.extend(str(stat) for stat in sorted_stats[:top_n])
        logger.info("\n".join(message))

    def log_object_memory_usage(
        self,
        logger: Logger,
        top_n: int,
        top_objects: List[tuple[Any, int]],
    ) -> None:
        message = [f"Top {top_n} objects in memory by size:"]
        for obj, size in top_objects:
            message.extend(f"{type(obj)}: {convert_bytes(size)}")
        logger.info("\n".join(message))


class MemoryTracer(MemoryTracerLogger):
    def __init__(self) -> None:
        self._name = type(self).__name__
        self._logger = module_logger(self._name)
        self._cached_snapshot: tracemalloc.Snapshot | None = None
        self._last_trace: datetime | None = None

    def start(self, top_n: int = 20, interval_seconds: int = 180) -> None:
        tracemalloc.start()

        t = threading.Thread(
            name=self._name,
            target=self.perform_memory_tracking,
            args=[top_n, interval_seconds],
        )
        t.start()

    def stop(self) -> None:
        tracemalloc.stop()

    def perform_memory_tracking(self, top_n: int, interval_seconds: int) -> None:
        while tracemalloc.is_tracing():
            if self._trace_is_due(interval_seconds):
                self._take_memory_snapshot(top_n)
                # self._dispatch_get_top_n_objects(top_n)
                self._last_trace = now_utc_iso()
            time.sleep(min(15, interval_seconds))

    def _trace_is_due(self, interval_seconds: int) -> bool:
        if self._last_trace is None:
            return True

        return now_utc_iso() - self._last_trace > timedelta(seconds=interval_seconds)

    def _take_memory_snapshot(self, top_n: int) -> None:
        snapshot = tracemalloc.take_snapshot()
        top_stats = None
        if self._cached_snapshot:
            top_stats = snapshot.compare_to(
                self._cached_snapshot,
                "lineno",
            )
        else:
            top_stats = snapshot.statistics("lineno")

        if top_stats:
            self.log_memory_allocations(self._logger, top_n, top_stats)

        self._cached_snapshot = snapshot

    # def _dispatch_get_top_n_objects(self, top_n: int) -> None:
    #     name = "object_size_tracking"
    #     if not find_thread_by_name(name):
    #         t = threading.Thread(
    #             target=self._get_top_n_objects,
    #             args=[top_n],
    #             name=name,
    #         )
    #         t.start()

    # def _get_top_n_objects(self, top_n: int) -> None:
    #     gc.collect()  # Clean up first
    #     objects = gc.get_objects()
    #     try:
    #         sizes: List[Tuple[Any, int]] = []
    #         for obj in objects:
    #             try:
    #                 size = asizeof(obj)
    #                 if len(sizes) < top_n:
    #                     sizes.append((obj, size))
    #                     sizes.sort(key=lambda x: x[1], reverse=True)
    #                 elif size > sizes[-1][1]:
    #                     # If the size is larger than the smallest size in the list, replace the smallest object
    #                     sizes[-1] = (obj, size)
    #                     sizes.sort(key=lambda x: x[1], reverse=True)
    #             except Exception:
    #                 continue
    #         top_objects = sorted(sizes, key=lambda x: x[1], reverse=True)[:top_n]
    #         self.log_object_memory_usage(self._logger, top_n, top_objects)
    #     except Exception:
    #         pass


# Test run
# from sir_utilities.logger import LogMode, module_logger, setup_logging

# print(sys.path)
# setup_logging(LogMode.DEBUG_PRINT_ONLY)
# mt = MemoryTracer()
# mt.start(interval_seconds=30)
# time.sleep(600)
# mt.stop()
